"""
Acquisition function used to find the next point to query in the active learning stage
    - If integer variables are
        - one-hot encoded,
            - frequency-based (hamming distance) are used for exploration
        - treated as continuous variable
            - max box method is used for exploration
    - The acquisition step may be splitted into multiple stages for different types of variables
        - may explore continuous, integer, and categorical variables in different steps
        - may explore them in one step

(C) 2021-2023 Mengjia Zhu, Alberto Bemporad
"""

import numpy as np
import pulp as plp


class active_learn:

    def __init__(self, prob):
        """
        Obtain and constructs all necessary attributes for self
        """
        self.lb = prob.lb
        self.ub = prob.ub
        self.nc = prob.nc
        self.nint = prob.nint
        self.nint_encoded = prob.nint_encoded
        self.nci = prob.nci
        self.nci_encoded = prob.nci_encoded
        self.int_encoded = prob.int_encoded
        self.int_interval = prob.int_interval
        self.nd = prob.nd
        self.X_d = prob.X_d
        self.sum_X_d = prob.sum_X_d
        self.nvars_encoded = prob.nvars_encoded
        self.maxevals = prob.maxevals

        self.scale_vars = prob.scale_vars
        self.dd_int = prob.dd_int
        self.d0_int = prob.d0_int

        self.delta_E = prob.delta_E
        self.timelimit = prob.timelimit

        self.isLin_eqConstrained = prob.isLin_eqConstrained
        self.Aeq = prob.Aeq
        self.beq = prob.beq
        self.isLin_ineqConstrained = prob.isLin_ineqConstrained
        self.Aineq = prob.Aineq
        self.bineq = prob.bineq


    def discrete_explore(self, X, zbest_nci, a, b, N, omega, gamma, Kf, dF):
        """
        Solve the acquisition step to obtain optimal categorical variables (encoded) based on current samples and surrogate model
            - treat the continuous variables, if exist, as constants
            - treat the integer variables, if exist, as constants
            - explore categorical variables

        Inputs:
                X: np array, training samples
                zbest_nci: np array, continuous/integer variables of the current best x
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xd_opt: np array, the optimum categorical variables

        """
        nc = self.nc
        nint = self.nint
        # nci = self.nci
        nci_encoded = self.nci_encoded
        # int_encoded = self.int_encoded
        nd = self.nd
        X_d = self.X_d
        sum_X_d = self.sum_X_d

        scale_vars = self.scale_vars
        dd_int = self.dd_int
        d0_int = self.d0_int

        zbest_c_unscaled = zbest_nci.copy()
        if (not self.int_encoded) and (scale_vars) and (nint >0):
            zbest_c_unscaled[nc:, 0] = zbest_nci[nc:, 0] * (np.ones((1, 1)) * dd_int) + np.ones((1, 1)) * d0_int

        delta_E = self.delta_E
        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('discrete_explore', plp.LpMinimize)
        xd = plp.LpVariable.dicts("xd", (range(sum_X_d), range(1)), cat=plp.LpBinary)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)
        cat_ham_dist_scaled = plp.LpVariable.dicts("cat_ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        cost = - delta_E * plp.lpSum(cat_ham_dist_scaled) + plp.lpSum(y) / dF
        prob += cost

        # constraints for one-hot encoded xd
        for i in range(nd):
            prob += (plp.lpSum(xd[h][0] for h in range(sum(X_d[:i]), sum(X_d[:i + 1]))) == 1)

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * zbest_c_unscaled[h][0] for h in range(nci_encoded))
                         + (plp.lpSum(Aeq[i, nci_encoded + h] * xd[h][0] for h in range(sum_X_d)))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * zbest_c_unscaled[h][0] for h in range(nci_encoded))
                         + (plp.lpSum(Aineq[i, nci_encoded + h] * xd[h][0] for h in range(sum_X_d)))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * zbest_nci[h][0] for h in range(nci_encoded))
                             + plp.lpSum((omega[i, nci_encoded + h] - omega[j, nci_encoded + h]) * xd[h][0] for h in range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = 1.0e5 * np.ones((Kf, 1))
        Mcm = -1.0e5 * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = plp.lpSum(a[j][h, 0] * zbest_nci[h][0] for h in range(nci_encoded)) + plp.lpSum(
                a[j][nci_encoded + h, 0] * xd[h][0] for h in range(sum_X_d))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))

        # acquisition function related
        sum_ham_dist = 0
        for i in range(N):
            ind_0 = list(np.where(X[i, :] == 0)[0])
            ind_1 = list(np.where(X[i, :] == 1)[0])
            sum_ham_dist += plp.lpSum(xd[h][0] for h in ind_0) + plp.lpSum(1 - xd[h][0] for h in ind_1)

        prob += cat_ham_dist_scaled == 1 / (nd * N) * sum_ham_dist

        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xd_opt = np.zeros((sum_X_d,))
            for i in range(sum_X_d):
                xd_opt[i,] = xd[i][0].varValue
        else:
            xd_opt = np.nan

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # cat_ham_dist_scaled = cat_ham_dist_scaled[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * cat_ham_dist_scaled + y_opt / dF
            # # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xd_opt


    def integ_explore(self, X, zbest_nc, zbest_d, a, b, N, omega, gamma, Kf, dF):
        """
        When integer variables are not one-hot encoded (max box)
        Solve the acquisition step to obtain optimal integer variables based on current samples and surrogate model
            - treat the categorical variables, if exist, as constants
            - treat the continuous variables, if exist, as constants
            - explore integer variables

        Inputs:
                X: np array, training samples
                zbest_nc: np array, continuous variables of the current best x
                zbest_d: np array, categorical variables of the current best x
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xc_opt: np array, the optimum continuous/integer variables

        """
        lb = self.lb
        ub = self.ub
        nc = self.nc
        nint = self.nint
        nci = self.nci
        sum_X_d = self.sum_X_d


        scale_vars = self.scale_vars
        dd_int = self.dd_int
        d0_int = self.d0_int

        delta_E = self.delta_E
        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('integ_explore', plp.LpMinimize)
        xint_c = plp.LpVariable.dicts("xint_c", (range(nint), range(1)), cat=plp.LpContinuous)
        xint = plp.LpVariable.dicts("xint", (range(nint), range(1)), cat=plp.LpInteger)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)
        if nint >0:
            if (N * nint < 120):
                delta_p_int = plp.LpVariable.dicts("delta_p_int", (range(N), range(nint)), cat=plp.LpBinary)
                delta_m_int = plp.LpVariable.dicts("delta_m_int", (range(N), range(nint)), cat=plp.LpBinary)
            else:
                delta_p_int = plp.LpVariable.dicts("delta_p_int", (range(20), range(nint)), cat=plp.LpBinary)
                delta_m_int = plp.LpVariable.dicts("delta_m_int", (range(20), range(nint)), cat=plp.LpBinary)
        beta_int = plp.LpVariable.dicts("beta_int", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        cost = - delta_E * (plp.lpSum(beta_int)) + plp.lpSum(y) / dF
        prob += cost

        # up and lower bound of xint
        for h in range(nint):
            prob += xint[h][0] <= ub[nc + h]
            prob += xint[h][0] >= lb[nc + h]

        # relationship between xint and xint_scaled, i.e., xint_c
        for h in range(nint):
            if scale_vars:
                prob += xint_c[h][0] * dd_int[h] + d0_int[h] == xint[h][0]
            else:
                prob += xint_c[h][0] == xint[h][0]

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * zbest_nc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint))
                         + (plp.lpSum(Aeq[i, nci + h] * zbest_d[h][0] for h in range(sum_X_d)))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * zbest_nc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint))
                         + (plp.lpSum(Aineq[i, nci + h] * zbest_d[h][0] for h in range(sum_X_d)))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * zbest_nc[h][0] for h in range(nc))
                             + plp.lpSum((omega[i, nc + h] - omega[j, nc + h]) * xint_c[h][0] for h in range(nint))
                             + plp.lpSum((omega[i, nci + h] - omega[j, nci + h]) * zbest_d[h][0] for h in range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = max(1.0e5,dF*10) * np.ones((Kf, 1))
        Mcm = -max(1.0e5,dF*10) * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = (plp.lpSum(a[j][h, 0] * zbest_nc[h][0] for h in range(nc))
                  + plp.lpSum(a[j][nc + h, 0] * xint_c[h][0] for h in range(nint))
                  + plp.lpSum(a[j][nci + h, 0] * zbest_d[h][0] for h in range(sum_X_d)))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))

        # acquisition function related

        # big-M for exploration function
        M_x = 1.0e5 * np.ones((1, nint))
        # M_x = np.zeros((1, nci))
        # max_X_curr = np.zeros((1, nci))
        # min_X_curr = np.zeros((1, nci))
        # for h in range(nci):
        #     max_X_curr[0][h] = np.amax(X[:, h])
        #     min_X_curr[0][h] = np.amin(X[:, h])
        # for i in range(nci):
        #     if max_X_curr[0,i] > 0:
        #         M_x[0,i] = 2 * max_X_curr[0,i] - min_X_curr[0,i]
        #     else:
        #         M_x[0,i] = max_X_curr[0,i] - 2 * min_X_curr[0,i]

        if nint >0:
            if (N * nint < 120):
                for i in range(N):
                    for j in range(nint):
                        prob += delta_p_int[i][j] <= 1 - delta_m_int[i][j]
                        prob += (xint_c[j][0] - X[i, j])*dd_int[j] >= beta_int[0] - M_x[0, j] * (1 - delta_p_int[i][j])
                        prob += (-xint_c[j][0] + X[i, j])*dd_int[j] >= beta_int[0] - M_x[0, j] * (1 - delta_m_int[i][j])
                    prob += (plp.lpSum(delta_p_int[i][j] + delta_m_int[i][j] for j in range(nint)) >= 1)
            else:
                N_iter = N - 20
                for i in range(20):
                    for j in range(nint):
                        prob += delta_p_int[i][j] <= 1 - delta_m_int[i][j]
                        prob += (xint_c[j][0] - X[N_iter+i, j]) * dd_int[j] >= beta_int[0] - M_x[0, j] * (1 - delta_p_int[i][j])
                        prob += (-xint_c[j][0] + X[N_iter+i, j]) * dd_int[j] >= beta_int[0] - M_x[0, j] * (1 - delta_m_int[i][j])
                prob += (plp.lpSum(delta_p_int[i][j] + delta_m_int[i][j] for j in range(nint)) >= 1)
        else:
            prob += beta_int[0] == 0

        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xint_c_opt = np.zeros((nint,))
            for i in range(nint):
                xint_c_opt[i,] = xint_c[i][0].varValue
        else:
            xint_c_opt = np.nan

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # beta_opt = beta[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * beta_opt + y_opt / dF
            # # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xint_c_opt


    def integ_explore_intEncoded(self, X, zbest_nc, zbest_d, a, b, N, omega, gamma, Kf, dF):
        """
        When integer variables are one-hot encoded (hamming distance)
        Solve the acquisition step to obtain optimal integer variables based on current samples and surrogate model
            - treat the categorical variables, if exist, as constants
            - treat the continuous variables, if exist, as constants
            - explore integer variables

        Inputs:
                X: np array, training samples
                zbest_nc: np array, continuous variables of the current best x
                zbest_d: np array, categorical variables of the current best x
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xc_opt: np array, the optimum continuous/integer variables

        """
        nc = self.nc
        nint = self.nint
        nci_encoded = self.nci_encoded
        nint_encoded = self.nint_encoded
        int_interval = self.int_interval
        sum_X_d = self.sum_X_d

        delta_E = self.delta_E
        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('integ_explore_intEncoded', plp.LpMinimize)
        xint = plp.LpVariable.dicts("xint", (range(nint_encoded), range(1)), cat=plp.LpBinary)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)
        int_ham_dist_scaled = plp.LpVariable.dicts("int_ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        cost = - delta_E * (plp.lpSum(int_ham_dist_scaled)) + plp.lpSum(y) / dF
        prob += cost

        # constraints for one-hot encoded xint
        for i in range(nint):
            prob += (plp.lpSum(
                xint[h][0] for h in range(int(round(sum(int_interval[:i]))), int(round(sum(int_interval[:i + 1]))))) == 1)

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * zbest_nc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + (plp.lpSum(Aeq[i, nci_encoded + h] * zbest_d[h][0] for h in range(sum_X_d)))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * zbest_nc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + (plp.lpSum(Aineq[i, nci_encoded + h] * zbest_d[h][0] for h in range(sum_X_d)))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * zbest_nc[h][0] for h in range(nc))
                             + plp.lpSum((omega[i, nc + h] - omega[j, nc + h]) * xint[h][0] for h in range(nint_encoded))
                             + plp.lpSum((omega[i, nci_encoded + h] - omega[j, nci_encoded + h]) * zbest_d[h][0] for h in range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = max(1.0e5,dF*10) * np.ones((Kf, 1))
        Mcm = -max(1.0e5,dF*10) * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = (plp.lpSum(a[j][h, 0] * zbest_nc[h][0] for h in range(nc))
                  + plp.lpSum(a[j][nc + h, 0] * xint[h][0] for h in range(nint_encoded))
                  + plp.lpSum(a[j][nci_encoded + h, 0] * zbest_d[h][0] for h in range(sum_X_d)))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))

        # acquisition function related
        int_sum_ham_dist = 0
        for i in range(N):
            ind_0 = list(np.where(X[i, :] == 0)[0])
            ind_1 = list(np.where(X[i, :] == 1)[0])
            int_sum_ham_dist += plp.lpSum(xint[h][0] for h in ind_0) + plp.lpSum(1 - xint[h][0] for h in ind_1)

        prob += int_ham_dist_scaled == 1 / (nint * N) * int_sum_ham_dist

        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xint_opt = np.zeros((nint_encoded,))
            for i in range(nint_encoded):
                xint_opt[i,] = xint[i][0].varValue
        else:
            xint_opt = np.nan

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # beta_opt = beta[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * beta_opt + y_opt / dF
            # # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xint_opt


    def cont_explore(self, X, zbest_int_c, zbest_d, a, b, N, omega, gamma, Kf, dF):
        """
        Solve the acquisition step to obtain optimal continuous variables based on current samples and surrogate model
            - treat the categorical variables, if exist, as constants
            - treat the integer variables, if exist, as constants
            - explore continuous variables

        Inputs:
                X: np array, training samples
                zbest_int_c: np array, integer variables of the current best x (scaled if not one-hot encoded and if scale_vars = True)
                zbest_d: np array, categorical variables of the current best x
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xc_opt: np array, the optimum continuous/integer variables

        """
        lb = self.lb
        ub = self.ub
        nc = self.nc
        nci_encoded = self.nci_encoded
        nint_encoded = self.nint_encoded
        sum_X_d = self.sum_X_d

        scale_vars = self.scale_vars
        dd_int = self.dd_int
        d0_int = self.d0_int

        zbest_int_c_unscaled = zbest_int_c.copy()
        if (not self.int_encoded) and (scale_vars) and (nint_encoded >0):
            zbest_int_c_unscaled[:,0] = zbest_int_c[:,0] * (np.ones((1, 1)) * dd_int) + np.ones((1, 1)) * d0_int

        delta_E = self.delta_E
        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('cont_explore', plp.LpMinimize)
        xc = plp.LpVariable.dicts("xc", (range(nc), range(1)), cat=plp.LpContinuous)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)
        if nc >0:
            if (N * nc < 120):
                delta_p = plp.LpVariable.dicts("delta_p", (range(N), range(nc)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(N), range(nc)), cat=plp.LpBinary)
            else:
                delta_p = plp.LpVariable.dicts("delta_p", (range(20), range(nc)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(20), range(nc)), cat=plp.LpBinary)
        beta = plp.LpVariable.dicts("beta", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        cost = - delta_E * (plp.lpSum(beta)) + plp.lpSum(y) / dF
        prob += cost

        # up and lower bound of xc
        for h in range(nc):
            prob += xc[h][0] <= ub[h]
            prob += xc[h][0] >= lb[h]

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * zbest_int_c_unscaled[h][0] for h in range(nint_encoded))
                         + (plp.lpSum(Aeq[i, nci_encoded + h] * zbest_d[h][0] for h in range(sum_X_d)))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * zbest_int_c_unscaled[h][0] for h in range(nint_encoded))
                         + (plp.lpSum(Aineq[i, nci_encoded + h] * zbest_d[h][0] for h in range(sum_X_d)))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * xc[h][0] for h in range(nc))
                             + plp.lpSum((omega[i, nc + h] - omega[j, nc + h]) * zbest_int_c[h][0] for h in range(nint_encoded))
                             + plp.lpSum((omega[i, nci_encoded + h] - omega[j, nci_encoded + h]) * zbest_d[h][0] for h in range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = max(1.0e5,dF*10) * np.ones((Kf, 1))
        Mcm = -max(1.0e5,dF*10) * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = (plp.lpSum(a[j][h, 0] * xc[h][0] for h in range(nc))
                  + plp.lpSum(a[j][nc + h, 0] * zbest_int_c[h][0] for h in range(nint_encoded))
                  + plp.lpSum(a[j][nci_encoded + h, 0] * zbest_d[h][0] for h in range(sum_X_d)))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))

        # acquisition function related

        # big-M for exploration function
        M_x = 1.0e5 * np.ones((1, nc))
        # M_x = np.zeros((1, nci))
        # max_X_curr = np.zeros((1, nci))
        # min_X_curr = np.zeros((1, nci))
        # for h in range(nci):
        #     max_X_curr[0][h] = np.amax(X[:, h])
        #     min_X_curr[0][h] = np.amin(X[:, h])
        # for i in range(nci):
        #     if max_X_curr[0,i] > 0:
        #         M_x[0,i] = 2 * max_X_curr[0,i] - min_X_curr[0,i]
        #     else:
        #         M_x[0,i] = max_X_curr[0,i] - 2 * min_X_curr[0,i]

        if nc > 0:
            if (N * nc < 120):
                for i in range(N):
                    for j in range(nc):
                        prob += delta_p[i][j] <= 1 - delta_m[i][j]
                        prob += xc[j][0] - X[i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X[i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nc)) >= 1)
            else:
                N_iter = N - 20
                for i in range(20):
                    for j in range(nc):
                        prob += delta_p[i][j] <= 1 - delta_m[i][j]
                        prob += xc[j][0] - X[N_iter + i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X[N_iter + i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nc)) >= 1)
        else:
            prob += beta[0] == 0

        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xc_opt = np.zeros((nc,))
            for i in range(nc):
                xc_opt[i,] = xc[i][0].varValue
        else:
            xc_opt = np.nan

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # beta_opt = beta[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * beta_opt + y_opt / dF
            # # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xc_opt


    def acq_explore(self, X, a, b, N, omega, gamma, Kf, dF):
        """
        When integer variables are NOT one-hot encoded
        Solve the acquisition step to obtain optimal point to query based on current samples and surrogate model
            - obtain the continuous, integer, and categorical variables, if exist, all in one-step
            - explore continuous, integer, and categorical variables

        Inputs:
                X: np array, training samples
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xopt: np array, the optimum variables (next point to query)

        """
        lb = self.lb
        ub = self.ub
        nc = self.nc
        nint = self.nint
        nci = self.nci
        nd = self.nd
        X_d = self.X_d
        sum_X_d = self.sum_X_d

        scale_vars = self.scale_vars
        dd_int = self.dd_int
        d0_int = self.d0_int

        X_curr_d = X[:, nci:]

        delta_E = self.delta_E
        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('acq_explore', plp.LpMinimize)
        xc = plp.LpVariable.dicts("xc", (range(nci), range(1)), cat=plp.LpContinuous)
        xd = plp.LpVariable.dicts("xd", (range(sum_X_d), range(1)), cat=plp.LpBinary)
        xint = plp.LpVariable.dicts("xint", (range(nint), range(1)), cat=plp.LpInteger)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)
        if nc >0:
            if (N * nc < 120) and (N * nint < 120):
                delta_p = plp.LpVariable.dicts("delta_p", (range(N), range(nci)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(N), range(nci)), cat=plp.LpBinary)
            else:
                delta_p = plp.LpVariable.dicts("delta_p", (range(20), range(nci)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(20), range(nci)), cat=plp.LpBinary)
        beta = plp.LpVariable.dicts("beta", range(1), lowBound=0, cat=plp.LpContinuous)
        if nint >0:
            if (N * nc < 120) and (N * nint < 120):
                delta_p_int = plp.LpVariable.dicts("delta_p_int", (range(N), range(nint)), cat=plp.LpBinary)
                delta_m_int = plp.LpVariable.dicts("delta_m_int", (range(N), range(nint)), cat=plp.LpBinary)
            else:
                delta_p_int = plp.LpVariable.dicts("delta_p_int", (range(20), range(nint)), cat=plp.LpBinary)
                delta_m_int = plp.LpVariable.dicts("delta_m_int", (range(20), range(nint)), cat=plp.LpBinary)
        beta_int = plp.LpVariable.dicts("beta_int", range(1), lowBound=0, cat=plp.LpContinuous)
        cat_ham_dist_scaled = plp.LpVariable.dicts("cat_ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        cost = - delta_E * (plp.lpSum(cat_ham_dist_scaled) + plp.lpSum(beta) + plp.lpSum(beta_int)) + plp.lpSum(y) / dF
        prob += cost

        # up and lower bound of xc
        for h in range(nc):
            prob += xc[h][0] <= ub[h]
            prob += xc[h][0] >= lb[h]

        # up and lower bound of xint
        for h in range(nint):
            prob += xint[h][0] <= ub[nc + h]
            prob += xint[h][0] >= lb[nc + h]

        # relationship between xint and xint_scaled
        for h in range(nint):
            if scale_vars:
                prob += xc[nc + h][0] * dd_int[h] + d0_int[h] == xint[h][0]
            else:
                prob += xc[nc + h][0] == xint[h][0]

        # constraints for one-hot encoded xd
        for i in range(nd):
            prob += (plp.lpSum(xd[h][0] for h in range(sum(X_d[:i]), sum(X_d[:i + 1]))) == 1)

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint))
                         + plp.lpSum(Aeq[i, nci + h] * xd[h][0] for h in range(sum_X_d))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint))
                         + plp.lpSum(Aineq[i, nci + h] * xd[h][0] for h in range(sum_X_d))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * xc[h][0] for h in range(nci))
                             + plp.lpSum((omega[i, nci + h] - omega[j, nci + h]) * xd[h][0] for h in range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = max(1.0e5,dF*10) * np.ones((Kf, 1))
        Mcm = -max(1.0e5,dF*10) * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = (plp.lpSum(a[j][h, 0] * xc[h][0] for h in range(nci))
                  + plp.lpSum(a[j][nci + h, 0] * xd[h][0] for h in range(sum_X_d)))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))

        # acquisition function related
        if nd > 0:
            # discrete variables
            sum_ham_dist = 0
            for i in range(N):
                ind_0 = list(np.where(X_curr_d[i, :] == 0)[0])
                ind_1 = list(np.where(X_curr_d[i, :] == 1)[0])
                sum_ham_dist += plp.lpSum(xd[h][0] for h in ind_0) + plp.lpSum(1 - xd[h][0] for h in ind_1)

            prob += cat_ham_dist_scaled == 1 / (nd * N) * sum_ham_dist
        else:
            prob += cat_ham_dist_scaled[0] == 0

        # continuous variables
        # big-M for exploration function
        M_x = 1.0e5 * np.ones((1, nci))

        # max_X_curr = np.zeros((1, nci))
        # min_X_curr = np.zeros((1, nci))
        # for h in range(nci):
        #     max_X_curr[0][h] = np.amax(X_curr_c[:, h])
        #     min_X_curr[0][h] = np.amin(X_curr_c[:, h])
        # for i in range(nci):
        #     if max_X_curr[0,i] > 0:
        #         M_x[0,i] = 2 * max_X_curr[0,i] - min_X_curr[0,i]
        #     else:
        #         M_x[0,i] = max_X_curr[0,i] - 2 * min_X_curr[0,i]

        if nc >0:
            if (N * nc < 120) and (N * nint < 120):
                for i in range(N):
                    for j in range(nc):
                        prob += delta_p[i][j] <= 1 - delta_m[i][j]
                        prob += xc[j][0] - X[i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X[i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nc)) >= 1)
            else:
                N_iter = N - 20
                for i in range(20):
                    for j in range(nc):
                        prob += delta_p[i][j] <= 1 - delta_m[i][j]
                        prob += xc[j][0] - X[N_iter + i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X[N_iter + i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nc)) >= 1)
        else:
            prob += beta[0] == 0

        if nint >0:
            if (N * nc < 120) and (N * nint < 120):
                for i in range(N):
                    for j in range(nint):
                        prob += delta_p_int[i][j] <= 1 - delta_m_int[i][j]
                        prob += (xc[nc+j][0] - X[i, nc+j])*dd_int[j] >= beta_int[0] - M_x[0, nc+j] * (1 - delta_p_int[i][j])
                        prob += (-xc[nc+j][0] + X[i, nc+j])*dd_int[j]  >= beta_int[0] - M_x[0, nc+j] * (1 - delta_m_int[i][j])
                    prob += (plp.lpSum(delta_p_int[i][j] + delta_m_int[i][j] for j in range(nint)) >= 1)
            else:
                N_iter = N - 20
                for i in range(20):
                    for j in range(nint):
                        prob += delta_p_int[i][j] <= 1 - delta_m_int[i][j]
                        prob += (xc[nc+j][0] - X[N_iter + i, nc+j])*dd_int[j] >= beta_int[0] - M_x[0, nc+j] * (1 - delta_p_int[i][j])
                        prob += (-xc[nc+j][0] + X[N_iter + i, nc+j])*dd_int[j]  >= beta_int[0] - M_x[0, nc+j] * (1 - delta_m_int[i][j])
                    prob += (plp.lpSum(delta_p_int[i][j] + delta_m_int[i][j] for j in range(nint)) >= 1)
        else:
            prob += beta_int[0] == 0


        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xopt = np.zeros((nci + sum_X_d,))
            for i in range(nci):
                xopt[i,] = xc[i][0].varValue
            for i in range(nci, nci + sum_X_d):
                xopt[i,] = xd[i - nci][0].varValue
        else:
            xopt = np.nan

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # cat_ham_dist_scaled = cat_ham_dist_scaled[0].varValue
            # beta_opt = beta[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * (cat_ham_dist_scaled + beta_opt) + y_opt / dF
            # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xopt


    def acq_explore_intEncoded(self, X, a, b, N, omega, gamma, Kf, dF):
        """
        When integer variables are one-hot encoded
        Solve the acquisition step to obtain optimal point to query based on current samples and surrogate model
            - obtain the continuous, integer, and categorical variables, if exist, all in one-step
            - explore continuous, integer, and categorical variables


        Inputs:
                X: np array, training samples
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xopt: np array, the optimum variables (next point to query)

        """
        lb = self.lb
        ub = self.ub
        nc = self.nc
        nint = self.nint
        nci_encoded = self.nci_encoded
        nint_encoded = self.nint_encoded
        int_interval = self.int_interval
        nd = self.nd
        X_d = self.X_d
        sum_X_d = self.sum_X_d

        # X_curr_c = X[:, :nc]
        X_curr_int = X[:, nc:nci_encoded]
        X_curr_d = X[:, nci_encoded:]

        delta_E = self.delta_E
        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('acq_explore_intEncoded', plp.LpMinimize)
        xc = plp.LpVariable.dicts("xc", (range(nc), range(1)), cat=plp.LpContinuous)
        xint = plp.LpVariable.dicts("xint", (range(nint_encoded), range(1)), cat=plp.LpBinary)
        xd = plp.LpVariable.dicts("xd", (range(sum_X_d), range(1)), cat=plp.LpBinary)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)
        if nc > 0:
            if (N * nc < 120):
                delta_p = plp.LpVariable.dicts("delta_p", (range(N), range(nc)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(N), range(nc)), cat=plp.LpBinary)
            else:
                delta_p = plp.LpVariable.dicts("delta_p", (range(20), range(nc)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(20), range(nc)), cat=plp.LpBinary)
        beta = plp.LpVariable.dicts("beta", range(1), lowBound=0, cat=plp.LpContinuous)
        int_ham_dist_scaled = plp.LpVariable.dicts("int_ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)
        cat_ham_dist_scaled = plp.LpVariable.dicts("cat_ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        cost = - delta_E * (plp.lpSum(cat_ham_dist_scaled) + plp.lpSum(beta) + plp.lpSum(int_ham_dist_scaled)) + plp.lpSum(y) / dF
        prob += cost

        # up and lower bound of xc
        for h in range(nc):
            prob += xc[h][0] <= ub[h]
            prob += xc[h][0] >= lb[h]

        # constraints for one-hot encoded xd
        for i in range(nd):
            prob += (plp.lpSum(xd[h][0] for h in range(sum(X_d[:i]), sum(X_d[:i + 1]))) == 1)

        # constraints for one-hot encoded xint
        for i in range(nint):
            prob += (plp.lpSum(xint[h][0] for h in range(int(round(sum(int_interval[:i]))), int(round(sum(int_interval[:i + 1]))))) == 1)


        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + plp.lpSum(Aeq[i, nci_encoded + h] * xd[h][0] for h in range(sum_X_d))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + plp.lpSum(Aineq[i, nci_encoded + h] * xd[h][0] for h in range(sum_X_d))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * xc[h][0] for h in range(nc))
                             + plp.lpSum((omega[i, nc+h] - omega[j, nc+h]) * xint[h][0] for h in range(nint_encoded))
                             + plp.lpSum((omega[i, nci_encoded + h] - omega[j, nci_encoded + h]) * xd[h][0] for h in range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = max(1.0e5,dF*10) * np.ones((Kf, 1))
        Mcm = -max(1.0e5,dF*10) * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = (plp.lpSum(a[j][h, 0] * xc[h][0] for h in range(nc))
                  + + plp.lpSum(a[j][nc + h, 0] * xint[h][0] for h in range(nint_encoded))
                  + plp.lpSum(a[j][nci_encoded + h, 0] * xd[h][0] for h in range(sum_X_d)))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))

        # acquisition function related
        if nd > 0:
            # discrete variables
            cat_sum_ham_dist = 0
            for i in range(N):
                ind_0 = list(np.where(X_curr_d[i, :] == 0)[0])
                ind_1 = list(np.where(X_curr_d[i, :] == 1)[0])
                cat_sum_ham_dist += plp.lpSum(xd[h][0] for h in ind_0) + plp.lpSum(1 - xd[h][0] for h in ind_1)

            prob += cat_ham_dist_scaled == 1 / (nd * N) * cat_sum_ham_dist
        else:
            prob += cat_ham_dist_scaled[0] == 0


        if nint > 0:
            # encoded integer variables
            int_sum_ham_dist = 0
            for i in range(N):
                ind_0 = list(np.where(X_curr_int[i, :] == 0)[0])
                ind_1 = list(np.where(X_curr_int[i, :] == 1)[0])
                int_sum_ham_dist += plp.lpSum(xint[h][0] for h in ind_0) + plp.lpSum(1 - xint[h][0] for h in ind_1)

            prob += int_ham_dist_scaled == 1 / (nd * N) * int_sum_ham_dist
        else:
            prob += int_ham_dist_scaled[0] == 0

        # continuous variables
        # big-M for exploration function
        M_x = 1.0e5 * np.ones((1, nc))

        # max_X_curr = np.zeros((1, nci))
        # min_X_curr = np.zeros((1, nci))
        # for h in range(nci):
        #     max_X_curr[0][h] = np.amax(X_curr_c[:, h])
        #     min_X_curr[0][h] = np.amin(X_curr_c[:, h])
        # for i in range(nci):
        #     if max_X_curr[0,i] > 0:
        #         M_x[0,i] = 2 * max_X_curr[0,i] - min_X_curr[0,i]
        #     else:
        #         M_x[0,i] = max_X_curr[0,i] - 2 * min_X_curr[0,i]

        if nc > 0:
            if (N * nc < 120):
                for i in range(N):
                    for j in range(nc):
                        prob += delta_p[i][j] <= 1 - delta_m[i][j]
                        prob += xc[j][0] - X[i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X[i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nc)) >= 1)
            else:
                N_iter = N - 20
                for i in range(20):
                    for j in range(nc):
                        prob += delta_p[i][j] <= 1 - delta_m[i][j]
                        prob += xc[j][0] - X[N_iter + i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X[N_iter + i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nc)) >= 1)
        else:
            prob += beta[0] == 0


        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xopt = np.zeros((nci_encoded + sum_X_d,))
            for i in range(nc):
                xopt[i,] = xc[i][0].varValue
            for i in range(nc, nci_encoded):
                xopt[i,] = xint[i - nc][0].varValue
            for i in range(nci_encoded, nci_encoded + sum_X_d):
                xopt[i,] = xd[i - nci_encoded][0].varValue
        else:
            xopt = np.nan

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # cat_ham_dist_scaled = cat_ham_dist_scaled[0].varValue
            # beta_opt = beta[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * (cat_ham_dist_scaled + beta_opt) + y_opt / dF
            # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xopt


    def acq_surrogate(self, a, b, omega, gamma, Kf, dF):
        """
        When integer variables are NOT one-hot encoded
        Solve the acquisition step to obtain optimal point to query based on the surrogate model
            - obtain the continuous, integer, and categorical variables, if exist, all in one-step
            - without exploration, based on the surrogate model ONLY

        Inputs:
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xopt: np array, the optimum variables (next point to query)

        """
        lb = self.lb
        ub = self.ub
        nc = self.nc
        nint = self.nint
        nci = self.nci
        nd = self.nd
        X_d = self.X_d
        sum_X_d = self.sum_X_d

        scale_vars = self.scale_vars
        dd_int = self.dd_int
        d0_int = self.d0_int

        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('Acq_exp', plp.LpMinimize)
        xc = plp.LpVariable.dicts("xc", (range(nci), range(1)), cat=plp.LpContinuous)
        xd = plp.LpVariable.dicts("xd", (range(sum_X_d), range(1)), cat=plp.LpBinary)
        xint = plp.LpVariable.dicts("xint", (range(nint), range(1)), cat=plp.LpInteger)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)

        # Objective function
        cost =  plp.lpSum(y) / dF
        prob += cost

        # up and lower bound of xc
        for h in range(nc):
            prob += xc[h][0] <= ub[h]
            prob += xc[h][0] >= lb[h]

        # up and lower bound of xint
        for h in range(nint):
            prob += xint[h][0] <= ub[nc + h]
            prob += xint[h][0] >= lb[nc + h]

        # relationship between xint and xint_scaled
        for h in range(nint):
            if scale_vars:
                prob += xc[nc + h][0] * dd_int[h] + d0_int[h] == xint[h][0]
            else:
                prob += xc[nc + h][0] == xint[h][0]

        # constraints for one-hot encoded xd
        for i in range(nd):
            prob += (plp.lpSum(xd[h][0] for h in range(sum(X_d[:i]), sum(X_d[:i + 1]))) == 1)

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint))
                         + plp.lpSum(Aeq[i, nci + h] * xd[h][0] for h in range(sum_X_d))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint))
                         + plp.lpSum(Aineq[i, nci + h] * xd[h][0] for h in range(sum_X_d))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * xc[h][0] for h in range(nci))
                             + plp.lpSum((omega[i, nci + h] - omega[j, nci + h]) * xd[h][0] for h in range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = max(1.0e5,dF*10) * np.ones((Kf, 1))
        Mcm = -max(1.0e5,dF*10) * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = (plp.lpSum(a[j][h, 0] * xc[h][0] for h in range(nci))
                  + plp.lpSum(a[j][nci + h, 0] * xd[h][0] for h in range(sum_X_d)))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))


        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xopt = np.zeros((nci + sum_X_d,))
            for i in range(nci):
                xopt[i,] = xc[i][0].varValue
            for i in range(nci, nci + sum_X_d):
                xopt[i,] = xd[i - nci][0].varValue

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # cat_ham_dist_scaled = cat_ham_dist_scaled[0].varValue
            # beta_opt = beta[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * (cat_ham_dist_scaled + beta_opt) + y_opt / dF
            # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xopt


    def acq_surrogate_intEncoded(self, a, b, omega, gamma, Kf, dF):
        """
        When integer variables are one-hot encoded
        Solve the acquisition step to obtain optimal point to query based on the surrogate model
            - obtain the continuous, integer, and categorical variables, if exist, all in one-step
            - without exploration, based on the surrogate model ONLY

        Inputs:
                a: np array, coefficients of the PWA
                b: np array, coefficients of the PWA
                N: int, number of training samples
                omega: np array, coefficients of the PWL
                gamma: np array, coefficients of the PWL
                Kf: number of partitions
                dF: float, scale parameter, defined as: max(max(y_pred)-min(y_pred), self.prob.epsDeltaF)

        Outputs:
                xopt: np array, the optimum variables (next point to query)

        """
        lb = self.lb
        ub = self.ub
        nc = self.nc
        nint = self.nint
        nint_encoded = self.nint_encoded
        int_interval = self.int_interval
        nci_encoded = self.nci_encoded
        nd = self.nd
        X_d = self.X_d
        sum_X_d = self.sum_X_d

        timelimit = self.timelimit

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        prob = plp.LpProblem('Acq_exp', plp.LpMinimize)
        xc = plp.LpVariable.dicts("xc", (range(nc), range(1)), cat=plp.LpContinuous)
        xint = plp.LpVariable.dicts("xint", (range(nint_encoded), range(1)), cat=plp.LpBinary)
        xd = plp.LpVariable.dicts("xd", (range(sum_X_d), range(1)), cat=plp.LpBinary)
        y = plp.LpVariable.dicts("y", range(1), cat=plp.LpContinuous)
        z_pwl_x = plp.LpVariable.dicts("z_pwl_x", range(Kf), cat=plp.LpBinary)
        v_pwl_x = plp.LpVariable.dicts("v_pwl_x", range(Kf), cat=plp.LpContinuous)

        # Objective function
        cost =  plp.lpSum(y) / dF
        prob += cost

        # up and lower bound of xc
        for h in range(nc):
            prob += xc[h][0] <= ub[h]
            prob += xc[h][0] >= lb[h]

        # constraints for one-hot encoded xd
        for i in range(nd):
            prob += (plp.lpSum(xd[h][0] for h in range(sum(X_d[:i]), sum(X_d[:i + 1]))) == 1)

        # constraints for one-hot encoded xint
        for i in range(nint):
            prob += (plp.lpSum(xint[h][0] for h in range(int(round(sum(int_interval[:i]))), int(round(sum(int_interval[:i + 1]))))) == 1)

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + plp.lpSum(Aeq[i, nci_encoded + h] * xd[h][0] for h in range(sum_X_d))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + plp.lpSum(Aineq[i, nci_encoded + h] * xd[h][0] for h in range(sum_X_d))
                         <= bineq[i][0])

        # Compute big-M for partition
        M = 1.0e5 * np.ones((Kf, Kf))
        # for j in range(Kf):
        #     for i in range(Kf):
        #         if not i == j:
        #             M[j, i] = sum((max(omega[i, h] - omega[j, h], 0) * ub[h]
        #                            - max(-omega[i, h] + omega[j, h], 0) * lb[h])
        #                           for h in range(nvar)) - gamma[j] + gamma[i]

        # Exclusive-or constraint
        prob += (plp.lpSum(z_pwl_x[i] for i in range(Kf)) == 1)

        # big-M constraint for PWL partition
        for j in range(Kf):
            for i in range(Kf):
                if not i == j:
                    prob += (plp.lpSum((omega[i, h] - omega[j, h]) * xc[h][0] for h in range(nc))
                             + plp.lpSum(
                                (omega[i, nc + h] - omega[j, nc + h]) * xint[h][0] for h in range(nint_encoded))
                             + plp.lpSum((omega[i, nci_encoded + h] - omega[j, nci_encoded + h]) * xd[h][0] for h in
                                         range(sum_X_d))
                             <= gamma[j] - gamma[i] + M[j, i] * (1 - z_pwl_x[j]))

        # Compute big-M for y
        Mcp = max(1.0e5,dF*10) * np.ones((Kf, 1))
        Mcm = -max(1.0e5,dF*10) * np.ones((Kf, 1))

        # for j in range(Kf):
        #     Mcp[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * ub
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * lb)) + b[j, 0]
        #     Mcm[j, 0] = np.sum(a[j][:, 0] * (a[j][:, 0] >= 0) * lb
        #                        - (-a[j][:, 0] * (a[j][:, 0] <= 0) * ub)) + b[j, 0]

        # big-M constraint for y
        for j in range(Kf):
            ax = (plp.lpSum(a[j][h, 0] * xc[h][0] for h in range(nc))
                  + + plp.lpSum(a[j][nc + h, 0] * xint[h][0] for h in range(nint_encoded))
                  + plp.lpSum(a[j][nci_encoded + h, 0] * xd[h][0] for h in range(sum_X_d)))
            prob += (v_pwl_x[j] <= ax + b[j] - Mcm[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] >= ax + b[j] - Mcp[j, 0] * (1 - z_pwl_x[j]))
            prob += (v_pwl_x[j] <= Mcp[j, 0] * z_pwl_x[j])
            prob += (v_pwl_x[j] >= Mcm[j, 0] * z_pwl_x[j])

        # Set y
        prob += (y == plp.lpSum(v_pwl_x[j] for j in range(Kf)))


        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xopt = np.zeros((nci_encoded + sum_X_d,))
            for i in range(nc):
                xopt[i,] = xc[i][0].varValue
            for i in range(nc, nci_encoded):
                xopt[i,] = xint[i - nc][0].varValue
            for i in range(nci_encoded, nci_encoded + sum_X_d):
                xopt[i,] = xd[i - nci_encoded][0].varValue

            # # to assess the results:
            # z_pwl_x_opt = np.zeros((Kf))
            # for i in range(Kf):
            #     z_pwl_x_opt[i,] = z_pwl_x[i].varValue
            # cat_ham_dist_scaled = cat_ham_dist_scaled[0].varValue
            # beta_opt = beta[0].varValue
            # y_opt = y[0].varValue
            # cost_opt = - delta_E * (cat_ham_dist_scaled + beta_opt) + y_opt / dF
            # cost_opt = - delta_E * beta_opt + y_opt

        # status_msg = plp.LpStatus[status]

        return xopt


"""
Fit the surrogate and obtain the relevant coefficients
    - Coefficient for PWL separation: omega, gamma
    - Coefficient for PWA in each partition: a, b

(C) 2021-2023 Mengjia Zhu, Alberto Bemporad
"""
import numpy as np

from sklearn.cluster import KMeans
from sklearn.linear_model import LogisticRegression
from scipy.special import logsumexp
import pulp as plp

import sys


class fit_surrogate:

    def __init__(self, prob):
        """
        Obtain and constructs all necessary attributes for self
        """
        self.nvars_encoded = prob.nvars_encoded
        self.timelimit = prob.timelimit
        self.sepvalue = prob.sepvalue
        self.K = prob.K
        self.initialization = prob.initialization
        self.separation = prob.separation
        self.K = prob.K
        self.nx = prob.nvars_encoded
        self.sigma = prob.sigma
        self.cost_tol = prob.cost_tol
        self.maxiter = prob.maxiter
        self.beta = prob.beta
        self.softmax_solver = prob.softmax_solver
        self.softmax_maxiter = prob.softmax_maxiter
        self.min_number = prob.min_number
        self.fit_on_partition = prob.fit_on_partition


    def get_parameters(self, X, I, Ieq, N, K, z_pwl_N):
        """
        Given the training samples, the corresponding preference relations,
         and the cluster assignments, determine the PWA parameters (a,b)

        Inputs:
                X: np array, the training samples
                I: np array, the preference relationship, I[i,1:2]=[h k] if F(h)<F(k)-comparetol
                Ieq: np array, the preference relationship, Ieq[i,1:2]=[h k] if |F(h)-F(k)|<=comparetol
                N: int, number of training samples
                K: int, number of partitions
                z_pwl_N:  np array with int elements, the assignments of each sampling point to the cluster

        Outputs:
                a_opt: np array, optimum coefficient a of PWA
                b_opt: np array, optimum coefficient b of PWA
                y_pred_opt: np array, the predicted function evaluations for the training samples X with the optimum coefficients
        """

        nvars_encoded = self.nvars_encoded
        timelimit = self.timelimit
        sepvalue = self.sepvalue

        m = I.shape[0]
        meq = Ieq.shape[0]
        # m_total = m + 2 * meq  # total number of comparisons
        m_total = m + meq

        prob_param = plp.LpProblem('PWASp_param', plp.LpMinimize)  # problem to find the parameters

        y_pred = plp.LpVariable.dicts("y_pred", range(N), cat=plp.LpContinuous)
        a = plp.LpVariable.dicts('a', (range(K), range(nvars_encoded)), cat=plp.LpContinuous)
        b = plp.LpVariable.dicts('b', range(K), cat=plp.LpContinuous)
        eps = plp.LpVariable.dicts("eps", range(m_total), lowBound=0, cat=plp.LpContinuous)
        reg_para = plp.LpVariable.dicts("reg_para", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        cost = plp.lpSum(eps) + 1.0e-6*plp.lpSum(reg_para)
        prob_param += cost

        # Exclusive-or constraint
        for n in range(N):
            # prob_param += (plp.lpSum(z_pwl_N[n][i] for i in range(K)) == 1.0)

            for j in range(K):
                if z_pwl_N[n] == j:
                    ax = plp.lpSum(a[j][h] * X[n][h] for h in range(nvars_encoded))
                    prob_param += (y_pred[n] == ax + b[j])


        # Objective function: Minimize sum|eps|
        for k in range(0, m):
            i = I[k][0]
            j = I[k][1]
            prob_param += y_pred[i] + sepvalue <= y_pred[j] + eps[k]

        for k in range(0, meq):
            i = Ieq[k][0]
            j = Ieq[k][1]
            prob_param += y_pred[i] - y_pred[j] <= sepvalue + eps[m + k]
            prob_param += y_pred[j] - y_pred[i] <= sepvalue + eps[m + k]

        # infinite norm regularization of coefficients
        for j in range(K):
            prob_param += reg_para >= -b[j]
            prob_param += reg_para >= b[j]
            for i in range(nvars_encoded):
                prob_param += reg_para >= -a[j][i]
                prob_param += reg_para >= a[j][i]

        try:
            prob_param.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob_param.solve(plp.GLPK(timeLimit=timelimit, msg=0))

        status = prob_param.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            a_opt = np.zeros((K, nvars_encoded, 1))
            b_opt = np.zeros((K, 1))
            y_pred_opt = np.zeros((N, 1))

            for k in range(K):
                b_opt[k] = b[k].varValue
                for var in range(nvars_encoded):
                    a_opt[k, var, 0] = a[k][var].varValue
            for n in range(N):
                y_pred_opt[n, 0] = y_pred[n].varValue

        return a_opt, b_opt, y_pred_opt


    def get_init_delta(self,X, N):
        """
        Get initial clustering of the training samples
            Reference code: 'parc_init()' function in PARC package by A. Bemporad, 2021
            use 'kmeans' or 'random' methods

        Inputs:
                X: np array, the training samples
                N: int, number of training samples

        Outputs:
                delta: np array with int elements, the initial assignments of each training samples to the clusters
        """

        K = self.K


        if K == 1:
            return np.zeros(N, dtype=int)

        raiseerror = False
        if not isinstance(self.initialization, tuple):
            raiseerror = True
        if (not raiseerror):
            if len(self.initialization) != 2:
                raiseerror = True
        if raiseerror:
            raise Exception('initialization option must be a tuple (string, value).')

        init_type = self.initialization[0]

        if init_type == "kmeans":
            # Run K-means++ to initialize delta
            kmeans = KMeans(n_clusters=K, init='k-means++',
                            n_init=self.initialization[1]).fit(X)
            delta = kmeans.labels_

        elif init_type == 'random':
            delta = np.random.random_integers(0, K - 1, N)  # random assignment
        else:
            raise Exception('Unknown initialization option "%s".' % init_type)

        return delta


    def get_pwl_param(self,X, delta, N):
        """
        Get optimum number of partitions, the cluster assignments, as well as the optimum coefficients for the PWL separation function
            reference code: PARC package by A. Bemporad, 2021

        Inputs:
                X: np array, the training samples
                delta: np array with int elements, the initial assignments of each training samples to the clusters
                N: int, number of training samples

        Outputs:
                Kf: int, optimum number of partitions
                delta: np array with int elements, the optimum assignments of each training samples to the clusters
                omega: np array, optimum coefficient a of the PWL separation function
                delta: np array, optimum coefficient a of the PWL separation function

        """

        separation = self.separation
        K = self.K
        nx = self.nvars_encoded
        sigma = self.sigma / N
        cost_tol = float(self.cost_tol)
        maxiter = int(self.maxiter)

        isSoftmax = (separation == 'Softmax')

        if not isSoftmax:
            errstr_softmax = 'Only \'Softmax\' separation method is implemented in this version. Please sepcify \'separation\' == \'Softmax\' '
            print(errstr_softmax)
            sys.exit(1)

        PWLsoftmax = LogisticRegression(multi_class='multinomial', C=2.0 / self.beta,
                                            solver=self.softmax_solver,
                                            max_iter=self.softmax_maxiter, tol=1e-6,
                                            warm_start=True)
        omega = np.zeros((K, nx))
        gamma = np.zeros(K)

        Nk = np.zeros(K, dtype=int)  # number of points in cluster
        killed = np.zeros(K, dtype=bool)  # clusters that have disappeared

        go = True
        iters = 0
        cost_old = np.inf
        cost_sequence = []

        while go:
            iters += 1
            ##########
            # Solve K ridge regression problems for the numeric targets
            ##########
            for j in range(K):
                if not killed[j]:
                    # Check if some clusters are empty
                    ii = (delta == j)
                    Nk[j] = np.sum(ii)
                    if Nk[j] == 0:
                        killed[j] = True

            ##########
            # find PWL separation function by softmax regression
            ##########
            if K > 1:
                omega1, gamma1 = self.fit_PWL(PWLsoftmax, X, delta)  # dim(gamma1) = unique(delta)
                h = 0
                for i in range(nx):
                    omega[~killed, i] = omega1[:, h]
                    h += 1
                gamma[~killed] = gamma1

            ##########
            # Compute cost and assign labels
            ##########
            cost = 0.0
            for k in range(N):
                cost_k = np.zeros(K)
                x_k = X[k, :].ravel()
                lin_terms = np.zeros(K)
                for j in range(K):
                    if not killed[j]:
                        aux = np.sum(omega[j, ] * x_k) + gamma[j]
                        cost_k[j] -= sigma * aux
                        lin_terms[j] = aux
                    else:
                        cost_k[j] = np.inf  # this prevents reassignement to killed cluster

                cost_k += sigma * logsumexp(lin_terms)

                # reassign labels
                delta[k] = np.argmin(cost_k)

                # compute current cost
                cost += cost_k[delta[k]]

            cost_sequence.append(cost)

            if (cost_old - cost <= cost_tol) or (iters == maxiter) or (K == 1):
                go = False

            cost_old = cost

        # update Nk and compute final number of clusters, possibly eliminating
        # empty or very small clusters, then clean-up solution
        for j in range(K):
            Nk[j] = np.sum(delta == j)
        killed = (Nk < self.min_number)

        isoutside = np.zeros(N, dtype=bool)
        anyoutside = False
        for i in range(K):
            if killed[i]:
                anyoutside = True
                isoutside[delta == i] = True

        NC = K - np.sum(killed)  # final number of clusters
        delta[isoutside] = -1  # mark points in small clusters with delta = -1
        Nk = np.zeros(NC, dtype=int)

        # recompute PWL partition based on final assignment
        for j in range(NC):
            ii = (delta == j)  # outliers are automatically excluded as j=0,...,NC-1
            Nk[j] = np.sum(ii)

        if NC > 1:
            if K > 1:
                omega = np.zeros((NC, nx))
                omega[:, ], gamma = self.fit_PWL(PWLsoftmax, X[~isoutside].T[:].T,
                                                 delta[~isoutside])  # dim(gamma1) = unique(delta[~isoutside])
        else:
            omega = np.zeros((1, nx))
            gamma = np.zeros(1)

        if not self.fit_on_partition:
            # re-label existing delta in (0,NC)
            elems = list(set(delta))
            if anyoutside:
                elems.remove(-1)
            for i in range(NC):
                delta[delta == elems[i]] = i
        else:
            # reassign points based on the polyhedron they belong to
            delta = np.argmax(X[:, ] @ omega[:, ].T + gamma, axis=1)
            delta[isoutside] = -1

            # some clusters may have disappeared after reassignment
            h = 0
            keep = np.ones(NC, dtype=bool)
            for j in range(NC):
                ii = (delta == j)
                aux = np.sum(ii)
                if aux >= self.min_number:
                    Nk[j] = aux
                    delta[ii] = h  # relabel points
                    h += 1
                else:
                    delta[ii] = -1  # also mark these as outliers, as they form an excessively small cluster
                    keep[j] = False


            if NC > h:
                omega = omega[keep, :]
                gamma = gamma[keep]
                Nk = Nk[keep]
            NC = h

            Kf = int(NC)

        return Kf, delta, omega, gamma


    def fit_PWL(self, softmax_reg, X, delta):
        """
        Fit a PWL separation function to the clusters obtained
            'fit_PWL' function taken from PARC package by A. Bemporad, 2021

        Inputs:
                softmax_reg: initial setup for the softmax regressor
                X: np array, the training samples
                delta: np array, the cluster assignments for each training sample

        Outputs:
                omega1: np array, coefficients of PWL separation function
                gamma1: np array, coefficients of PWL separation function

        """
        try:
            softmax_reg.fit(X, delta)
        except:
            if not len(np.unique(delta)) == 1:
                softmax_reg.warm_start = False  # new problem has different number of classes, disable warm start
                softmax_reg.fit(X, delta)
                softmax_reg.warm_start = True
        if len(np.unique(delta)) > 2:
            omega1 = softmax_reg.coef_
            gamma1 = softmax_reg.intercept_
        elif len(np.unique(delta)) == 1:
            omega1 = np.zeros((1, X.shape[1]))
            gamma1 = np.zeros(1)
        else:
            omega1 = np.zeros((2, X.shape[1]))
            gamma1 = np.zeros(2)
            omega1[0, :] = -softmax_reg.coef_
            gamma1[0] = -softmax_reg.intercept_
            omega1[1, :] = softmax_reg.coef_
            gamma1[1] = softmax_reg.intercept_

        return omega1, gamma1